start_time_model <- Sys.time()


#Load ref model to know the x,y of the Catalunya domain
# Extract domain ########################################################
netcdf_ref <- RNetCDF::open.nc("/gpfs/scratch/bsc32/bsc032454/data_datafusion/model_ref/sconcno2_2022010100.nc") #/esarchive/scratch/ccarnere/AQS_database/caliope/cat/hourly/sconcno2/sconcno2_2022010100.nc

x_ref <- RNetCDF::var.get.nc(netcdf_ref, 'x')
y_ref <- RNetCDF::var.get.nc(netcdf_ref, 'y')


# Load Model ############################################################
#This script is called from UK to set model data:
contador <- 0
empty_model <- NULL
model <- NULL
#set_parallel <- TRUE
#set_cpus <- 8 
sfStop()
sfInit(parallel=set_parallel, set_cpus, type="SOCK") 
sfLibrary(RNetCDF)
sfExport("netcdf_ref", "x_ref", "y_ref", "contador", "empty_model", "model_path", 
         "name_pollutant", "model_time_resolution", "mod_constant", "time_resolution")

# We download the model by days
model_list <- sfLapply(dates, function(i_date) {
  #i_date <- i_dates[1]
  
  conc_path_day <- paste0(i_date, "00")
  file_path <- paste0(model_path, name_pollutant, "_", conc_path_day, ".nc")
  
  if (file.exists(file_path)) {
    netcdf_model <- RNetCDF::open.nc(file_path)  #tancar els fitxer
    
    # crop model
    if(!exists("x_crop") || !exists("y_crop")) {
      y_model <- RNetCDF::var.get.nc(netcdf_model, 'y')
      x_model <- RNetCDF::var.get.nc(netcdf_model, 'x')
      
      id_max_y <- max(which(y_model <= max(y_ref)))
      id_min_y <- min(which(y_model >= min(y_ref)))
      id_max_x <- max(which(x_model <= max(x_ref)))
      id_min_x <- min(which(x_model >= min(x_ref)))
      
      y_crop <- id_min_y:id_max_y
      x_crop <- id_min_x:id_max_x
      
      # Make x_crop and y_crop available outside the function
      assign("x_crop", x_crop, envir = .GlobalEnv)
      assign("y_crop", y_crop, envir = .GlobalEnv)
    }
    
    # Extract model data
    if(substr(i_date, 1, 4) == 2017 & name_pollutant %in% c("pm10", "pm2p5")) {
      name_pollutant_var <- paste0(name_pollutant, "cmaq")
    } else {
      name_pollutant_var <- name_pollutant
    }
    
    model_day <- RNetCDF::var.get.nc(netcdf_model, name_pollutant_var) 
    
    # crop model
    if(model_time_resolution == "hourly") {
      model_day <- model_day[x_crop, y_crop, 1:24]
    } else if(model_time_resolution == "daily") {
      model_day <- model_day[x_crop, y_crop]
    } else {
      print("Wrong time resolution")
      return(NULL)
    }
    
    if(!exists("n_lat") || !exists("n_lon")) {
      n_lat <- dim(model_day)[1]
      n_lon <- dim(model_day)[2]
      
      # Make dimensions available outside the function
      assign("n_lat", n_lat, envir = .GlobalEnv)
      assign("n_lon", n_lon, envir = .GlobalEnv)
      
      # Handle missing days at beginning
      if(model_time_resolution == "hourly") {
        assign("model", array(dim = c(n_lat, n_lon, contador * 24)), envir = .GlobalEnv)
      } else if(model_time_resolution == "daily") {
        assign("model", array(dim = c(n_lat, n_lon, contador)), envir = .GlobalEnv)
      }
    }
    
    # Convert units if needed
    if(name_pollutant != "co") {
      model_day <- model_day * mod_constant 
    }
    
    # Handle time resolution differences
    if(model_time_resolution == "hourly" & time_resolution == "daily") {
      averaged_array <- apply(model_day, c(1, 2), mean)
      model_day <- array(averaged_array, dim = c(nrow(averaged_array), ncol(averaged_array), 1))
    } else if(model_time_resolution == "daily" & time_resolution == "hourly") {
      print(paste0("WARNING: time resolution of the model and other data (observations mostly), do not match"))
    }
    
  } else {
    # Handle missing data
    assign("contador", contador + 1, envir = .GlobalEnv)
    assign("empty_model", c(empty_model, i_date), envir = .GlobalEnv)
    print(paste0("WARNING: Model not found on ", i_date, ". Filling with NANs."))
    
    if(!exists("n_lat") || !exists("n_lon")) {
      return(NULL)
    } else {
      if(time_resolution == "hourly") {
        model_day <- array(NaN, dim = c(n_lat, n_lon, 24)) 
      } else if(time_resolution == "daily") {
        model_day <- array(NaN, dim = c(n_lat, n_lon, 1))
      } else {
        print("Wrong model_time_resolution")
        return(NULL)
      }
    }
    RNetCDF::close.nc(netcdf_model)
  }
  
  return(model_day)
})
sfStop()

#remove model_day when empty (de moment no ho poso, pq el q volem es omplirlo de nans i no hauria dhaverhi cap buit)
#model_list <- model_list[!sapply(model_list, is.null)] 

model <- do.call(abind, c(model_list, along=3))
# Combine all model days
#model <- do.call(function(...) abind(..., along = 3), c(list(model), model_list))

#removing model list
rm(model_list)
gc(reset = TRUE)  #forces a fer una neteja de  la RAM



# if we had days missing model from the first day of the year, we will have gaps
# we fix that by filling from the top with NANs
n_gaps <- length(dates) - dim(model)[3]  #days with gaps from the 1st day
if(time_resolution=="hourly"){
  n_gaps <- n_gaps*24
}  #hours with gaps from the 1st day
if(n_gaps>0){
  model_gap_top <- array(NaN, dim=c(n_lat,n_lon,n_gaps))
  
  model <- abind(model_gap_top, model, along=3)
  rm(model_gap_top)
}


print(paste0("The number of missing raw model (CALIOPE) is: ", contador))
if(!is.null(empty_model)){
  print("These days are: ")
  print(empty_model)
}


#-----------------to do: extraure un model de referencia 

# extract model coordinates, cropping if necessary ########################
lat_model <- RNetCDF::var.get.nc(netcdf_ref, 'lat')
lon_model <- RNetCDF::var.get.nc(netcdf_ref, 'lon')

lat_model <- lat_model[x_crop,y_crop]
lon_model <- lon_model[x_crop,y_crop]

lat_model <- sapply(lat_model, as.numeric)
lon_model <- sapply(lon_model, as.numeric)

coords_model <- cbind(lon_model, lat_model)


#extract the dates in which we have model #########################################
available_dates_model <- dates[!dates %in% empty_model]

if(time_resolution=="hourly"){
  available_dates_model <- rep(available_dates_model, 24) %>%
    sort() %>%
    paste(seq(0,23)) %>% 
    as.POSIXct(format="%Y%m%d %H")
}


# save #############################################################################
if (save_model){
  saveRDS(object=model, file=paste0(save_result,"model_",file_format, "_",pollutant,".rds"))
  saveRDS(object=coords_model, file=paste0(save_result,"coords_model_",file_format, "_",pollutant,".rds"))
  print("Model saved")
} else {
  print("Model have not been saved")
}

end_time_model    <- Sys.time()
temps_total_model <- end_time_model - start_time_model
print(paste0("local model, execution time :",temps_total_model, "s" ))
